import os  # for getting directories for C++ files
import shutil  # for deleting/copying files
import ast  # for dependency graphing
import numpy as np
from gillespy2.core import log, Species


"""
This file contains various functions used in the ssa_c_solver, variable_ssa_c_solver, and numpy solvers.


C SOLVER FUNCTIONS BELOW
"""


def find_time(array, value):
    """
    Finds the index of the closest value in the array parameter, to the value parameter
    :param array: results['time'] array, input to find index of closest 'value' parameter
    :type array: numpy.ndarray
    :param value: Value in which to find the closest values index in the array parameter.
    :type value: float
    :return: Integer index, the index of the closest value to 'value' parameter.
    """
    index = np.searchsorted(array, value, side="left")
    return index


def _copy_files(destination, GILLESPY_C_DIRECTORY):
    src_files = os.listdir(GILLESPY_C_DIRECTORY)
    for src_file in src_files:
        src_file = os.path.join(GILLESPY_C_DIRECTORY, src_file)
        if os.path.isfile(src_file):
            shutil.copy(src_file, destination)


def _write_propensity(outfile, model, species_mappings, parameter_mappings, reactions):
    """
    This functions writes a models propensity functions to a cpp user simulation template, for the SSACSolvers.
    :param outfile: File where the propensity function will be written to
    :param model: Model used to access species, reactions
    :param species_mappings: Sanitized species names
    :param parameter_mappings: Sanitized parameter names
    :param reactions: Names of reactions
    :type reactions: str
    """
    for i in range(len(reactions)):
        # Write switch statement case for reaction
        outfile.write("""
        case {0}:
            return {1};
        """.format(i, model.listOfReactions[reactions[i]].sanitized_propensity_function(species_mappings,
                                                                                        parameter_mappings)))


def _write_reactions(outfile, model, reactions, species):
    customrxns = {}
    for i in range(len(reactions)):
        reaction = model.listOfReactions[reactions[i]]
        if reaction.type == 'customized':
            customrxns[i] = species_parse(model, reaction.propensity_function)
        for j in range(len(species)):
            change = (reaction.products.get(model.listOfSpecies[species[j]], 0)) - (reaction.reactants.get
                                                                                    (model.listOfSpecies[species[j]], 0)
                                                                                    )
            if change != 0:
                outfile.write("model.reactions[{0}].species_change[{1}] = {2};\n".format(i, j, change))

    for i in customrxns.keys():
        for j in range(len(reactions)):
            if i == j:
                continue
            if any(elem in customrxns[i] for elem in list(model.listOfReactions[reactions[j]].reactants)) or \
                    any(elem in customrxns[i] for elem in list(model.listOfReactions[reactions[j]].products)):
                outfile.write("model.reactions[{0}].affected_reactions.push_back({1});\n".format(i, j))


def _parse_binary_output(results_buffer, number_of_trajectories, number_timesteps, number_species, data, pause=False):
    """
    This function reads binary output from a CPP simulation
    :param results_buffer: stdout of the CPP simulation ran
    :param number_of_trajectories: Total number of trajectories for a simulation
    :type number_of_trajectories: int
    :param number_timesteps: How many steps for a given simulation
    :type number_timesteps: int
    :param number_species: Total number of species in a model
    :param pause: Whether or not a model was paused, set to true when simulation was sent a KeyBoardInterrupt or
    timeout.
    :return: Trajectories for a simulation, and time that simulation was stopped, if sent a keyboardinterrupt or
    timeout.
    """
    trajectory_base = np.empty((number_of_trajectories, number_timesteps, number_species+1))

    # Timestopped is added to the end of the data, when a simulation completes or is paused
    if pause:
        timeStopped = int(data[-1])
        data.pop()
    else:
        timeStopped = 0
    for t in range(number_of_trajectories):
        for i in range(number_timesteps*(number_species+1)):
            index = i + (number_timesteps*(number_species+1)*t)
            trajectory_base[t][i//(number_species+1)][i % (number_species+1)] = data[index]

    return trajectory_base, timeStopped



def c_solver_resume(timeStopped, simulation_data, t, resume=None):
    """
    If a simulation is being resumed from a previous simulation, this function is called in the VariableSSACSolver,
    or SSACSolver
    :param timeStopped: The time that a simulation was stopped, originally attained from the results_buffer returned
    by the CPP simulation.
    :param simulation_data: The current simulation data, attained after parsing the results in the VariableSSACSolver or
    SSACSolver.
    :param t: The end time for the resume simulation, originally set in model.run(t=...)
    :param resume: The previous simulations data
    :type resume: gillespy2.core.result object
    :return: Combined data of the previous simulation, and the current simulation
    """

    # If simulation was paused/KeyboardInterrupt
    if timeStopped != 0:
        cutoff = find_time(simulation_data[0]['time'], timeStopped)
        if cutoff == 0 or cutoff == 1:
            log.warning('You have paused the simulation too early, and no points have been calculated past'
                        ' initial values. A graphic display will not produce expected results.')
        else:
            cutoff -= 1
        for i in simulation_data[0]:
            simulation_data[0][i] = simulation_data[0][i][:cutoff]

    if resume is not None:
        resumeTime = float(resume['time'][-1])
        step = resumeTime - resume['time'][-2]
        if timeStopped == 0:
            timeSpan = np.arange(resumeTime, t + resumeTime + step, step)
        else:
            timeSpan = np.arange(resumeTime + step, timeStopped + resumeTime + step, step)
        simulation_data[0]['time'] = timeSpan

    if resume is not None:
        # If resuming, combine old pause with new data, and delete any excess null data
        for i in simulation_data[0]:
            oldData = resume[i]
            newData = simulation_data[0][i]
            simulation_data[0][i] = np.concatenate((oldData, newData), axis=None)
            if len(simulation_data[0]['time']) != len(simulation_data[0][i]):
                simulation_data[0]['time'] = simulation_data[0]['time'][:-1]
    return simulation_data


"""
NUMPY SOLVER UTILITIES BELOW
"""


def numpy_initialization(model):
    species_mappings = model.sanitized_species_names()
    species = list(species_mappings.keys())
    parameter_mappings = model.sanitized_parameter_names()
    number_species = len(species)
    return species_mappings, species, parameter_mappings, number_species


def numpy_trajectory_base_initialization(model, number_of_trajectories, timeline, species, resume=None):
    trajectory_base = np.zeros((number_of_trajectories, timeline.size, len(species) + 1))

    # copy time values to all trajectory row starts
    trajectory_base[:, :, 0] = timeline
    tmpSpecies = {}
    # copy initial populations to base
    if resume is not None:
        # Set initial values of species to where last left off
        for i in species:
            tmpSpecies[i] = resume[i][-1]
        for i, s in enumerate(species):
            trajectory_base[:, 0, i + 1] = tmpSpecies[s]
    else:
        for i, s in enumerate(species):
            trajectory_base[:, 0, i + 1] = model.listOfSpecies[s].initial_value

    return trajectory_base, tmpSpecies


def numpy_resume(timeStopped, simulation_data, resume=None):
    """
    Helper function for when resuming a simulation in a numpy based solver
    :param timeStopped: The time in which the simulation was stopped.
    :param simulation_data: The current models simulation data, after being parsed in the numpy solver of choice.
    :param resume: The previous simulations data, that is being resumed
    :type resume: gillespy2.core.results object
    :return: Combined simulation data, the old resume data and the current simulation data.
    """
    if timeStopped != 0:
        if timeStopped != simulation_data[0]['time'][-1]:
            tester = np.where(simulation_data[0]['time'] > timeStopped)[0].size
            index = np.where(simulation_data[0]['time'] == timeStopped)[0][0]
        if tester > 0:
            for i in simulation_data[0]:
                simulation_data[0][i] = simulation_data[0][i][:index]
    if resume is not None:
        # If resuming, combine old pause with new data, and delete any excess null data
        for i in simulation_data[0]:
            oldData = resume[i][:-1]
            newData = simulation_data[0][i]
            simulation_data[0][i] = np.concatenate((oldData, newData), axis=None)

    return simulation_data


"""
Below are two functions used for creating dependency graphs in the C solvers, and Numpy Solvers.
"""


def species_parse(model, custom_prop_fun):
    """
    This function uses Pythons AST module to parse custom propensity function, looking for Species in a model
    :param model: Model to be checked for species
    :param custom_prop_fun: The custom propensity function to be parsed
    :return: List of species objects that are found in a custom propensity function
    """
    parsed_species = []

    class SpeciesParser(ast.NodeTransformer):
        def visit_Name(self, node):
            if isinstance(model.get_element(node.id), Species):
                parsed_species.append(model.get_element(node.id))

    expr = custom_prop_fun
    expr = ast.parse(expr, mode='eval')
    expr = SpeciesParser().visit(expr)
    return parsed_species


def dependency_grapher(model, reactions):
    """
    This function returns a dependency graph for a models reactions in the form of a
    dictionary containing {species name: {'dependencies'}:[list of reaction names]}.
    :param model: Model to used to create a reaction dependency graph
    :param reactions: list(model.listOfReactions)
    :return: Dependency graph dictionary
    """
    dependent_rxns = {}
    for i in reactions:
        cust_spec = []
        if model.listOfReactions[i].type == 'customized':
            cust_spec = (species_parse(model, model.listOfReactions[i].propensity_function))

        for j in reactions:

            if i not in dependent_rxns:
                dependent_rxns[i] = {'dependencies': []}
            if j not in dependent_rxns:
                dependent_rxns[j] = {'dependencies': []}
            if i == j:
                continue

            reactantsI = list(model.listOfReactions[i].reactants.keys())
            reactantsJ = list(model.listOfReactions[j].reactants.keys())

            if j not in dependent_rxns[i]['dependencies']:
                if any(elem in reactantsI for elem in reactantsJ):
                    if i not in dependent_rxns[j]['dependencies']:
                        dependent_rxns[j]['dependencies'].append(i)
                    dependent_rxns[i]['dependencies'].append(j)

            if i not in dependent_rxns[j]['dependencies']:
                if any(elem in list(model.listOfReactions[i].products.keys()) for elem in
                       list(model.listOfReactions[j].reactants.keys())):
                    dependent_rxns[j]['dependencies'].append(i)

            if cust_spec:
                if any(elem in cust_spec for elem in list(model.listOfReactions[j].reactants)) or any \
                            (elem in cust_spec for elem in list(model.listOfReactions[j].products)):
                    dependent_rxns[i]['dependencies'].append(j)

    return dependent_rxns




import unittest
from example_models import Example
from gillespy2 import NumPySSASolver


class TestNumPySSASolver(unittest.TestCase):
    model = Example()
    

if __name__ == '__main__':
    unittest.main()



import unittest
from gillespy2.solvers.utilities.solverutils import dependency_grapher
from gillespy2.core import Reaction

r1 = Reaction(name='r1', propensity_function="5*x^2+e*b+6")
r2 = Reaction(name='r2', propensity_function="5*x**2+e*b+6")

r3 = Reaction(name='r3', propensity_function="1*alpha/2+5^beta")
r4 = Reaction(name='r4', propensity_function="1*alpha/2+5**beta")

r5 = Reaction(name='r5', propensity_function="2.78*x+3^(4*x)")
r6 = Reaction(name='r6', propensity_function="2.78*x+3**(4*x)")

r7 = Reaction(name='r7', propensity_function="(alpha/beta + delta**gamma)/(atlas-zeta)")
r8 = Reaction(name='r8', propensity_function="(alpha/beta + delta^gamma)/(atlas-zeta)")

r9 = Reaction(name='r9', propensity_function="-5*-x^2")
r10 = Reaction(name='r10', propensity_function="-5*-x**2")


class TestPropensityFunctions(unittest.TestCase):
    def test_propensity_functions(self):
        self.assertEqual(r1.propensity_function, "(((5*pow(x,2))+(2.718281828459045*b))+6)",
                         msg="Has incorrect expression")
        self.assertEqual(r2.propensity_function, "(((5*pow(x,2))+(2.718281828459045*b))+6)",
                         msg="Has incorrect expression")

        self.assertEqual(r3.propensity_function, "(((1*alpha)/2)+pow(5,beta))", msg="Has incorrect expression")
        self.assertEqual(r4.propensity_function, "(((1*alpha)/2)+pow(5,beta))", msg="Has incorrect expression")

        self.assertEqual(r5.propensity_function, "((2.78*x)+pow(3,(4*x)))", msg="Has incorrect expression")
        self.assertEqual(r6.propensity_function, "((2.78*x)+pow(3,(4*x)))", msg="Has incorrect expression")

        self.assertEqual(r7.propensity_function, "(((alpha/beta)+pow(delta,gamma))/(atlas-zeta))",
                         msg="Has incorrect expression")
        self.assertEqual(r8.propensity_function, "(((alpha/beta)+pow(delta,gamma))/(atlas-zeta))",
                         msg="Has incorrect expression")

        self.assertEqual(r9.propensity_function, "((-5)*(-pow(x,2)))",
                         msg="Has incorrect expression")
        self.assertEqual(r10.propensity_function, "((-5)*(-pow(x,2)))",
                         msg="Has incorrect expression")

    def test_dependency_graphing(self):
        from example_models import ToggleSwitch, MichaelisMenten
        model = ToggleSwitch()
        dependencies = dependency_grapher(model, list(model.listOfReactions.keys()))
        correct_graph = {'cu': {'dependencies': ['cv', 'dv']}, 'cv': {'dependencies': ['cu', 'du']},
                         'du': {'dependencies': ['cu']}, 'dv': {'dependencies': ['cv']}}
        self.assertEqual(correct_graph, dependencies)

        model = MichaelisMenten()
        dependencies = dependency_grapher(model, list(model.listOfReactions.keys()))
        correct_graph = {'r1': {'dependencies': ['r2', 'r3']}, 'r2': {'dependencies': ['r1', 'r3']},
                         'r3': {'dependencies': ['r1', 'r2']}}
        self.assertEqual(correct_graph, dependencies)




"""
This file contains a function and variable for testing a machines support of GillesPy2 C++ solvers.
Used in model.py
"""


def check_cpp_support():
    from gillespy2.solvers.cpp.example_models import Example
    from gillespy2 import SSACSolver
    try:
        model = Example()
        results = model.run(solver=SSACSolver, cpp_support=True)
        return True
    except Exception as e:
        log.warn('Unable to use C++ optimized SSA: {0}.  The performance of ' \
        'this package can be significantly increased if you install/configure GCC on ' \
        'this machine.'.format(e))
        return False


cpp_support = check_cpp_support()



from gillespy2.core.sortableobject import SortableObject

class FunctionDefinition(SortableObject):
    """
    Object representation defining an evaluable function to be used during
    simulation of a GillesPy2 model

    :param name: Name of the function to be made and called
    :type name: str
    :param function: Defined function body of operation to be performed.
    :type function: str
    :param variables: String names of Variables to be used as arguments to function.
    :type variables: list of strings
    """

    def __init__(self, name="", function=None, args=[]):

        import math
        eval_globals = math.__dict__

        self.name = name
        args = ', '.join(args)
        self.function = eval('lambda ' + args + ': ' + function, eval_globals)
        if self.function is None:
            raise TypeError

    def sanitized_function(self, species_mappings, parameter_mappings):
        names = sorted(list(species_mappings.keys()) + list(parameter_mappings.keys()), key=lambda x: len(x),
                       reverse=True)
        replacements = [parameter_mappings[name] if name in parameter_mappings else species_mappings[name]
                        for name in names]
        sanitized_function = self.function
        for id, name in enumerate(names):
            sanitized_function = sanitized_function.replace(name, "{" + str(id) + "}")
        return sanitized_function.format(*replacements)


from gillespy2.core.gillespyError import ExecutionError
from .cpp import *
from .numpy import *

__all__ = cpp.__all__ + numpy.__all__

if not __all__:
    raise ExecutionError("Your computer does not contain the minimum require"
                         "ments for running simulations using GillesPy2."
                         " Please install NumPy, or configure the C++ 'g++' "
                         "compiler on your machine.")



from gillespy2.core.sortableobject import SortableObject
from gillespy2.core.gillespyError import *
import numpy as np
import uuid
import ast


class Reaction(SortableObject):
    """
    Models a single reaction. A reaction has its own dicts of species
    (reactants and products) and parameters. The reaction's propensity
    function needs to be evaluable (and result in a non-negative scalar
    value) in the namespace defined by the union of those dicts.

    :param name: The name by which the reaction is called (optional).
    :type name: str
    :param reactants: The reactants that are consumed in the reaction, with stoichiometry. An
    example would be {R1 : 1, R2 : 2} if the reaction consumes two of R1 and
    one of R2, where R1 and R2 are Species objects.
    :type reactants: dict
    :param products: The species that are created by the reaction event, with stoichiometry. Same format as reactants.
    :type products: dict
    :param propensity_function: The custom propensity function for the reaction. Must be evaluable in the
    namespace of the reaction using C operations.
    :type propensity_function: str
    :param massaction: The switch to use a mass-action reaction. If set to True, a rate value is required.
    :type massaction: bool
    :param rate: The rate of the mass-action reaction, take care to note the units.
    :type rate: float
    :param annotation: An optional note about the reaction.
    :type annotation: str

    Notes
    ----------
    For a species that is NOT consumed in the reaction but is part of a mass
    action reaction, add it as both a reactant and a product.

    Mass-action reactions must also have a rate term added. Note that the input
    rate represents the mass-action constant rate independent of volume.

    if a name is not provided for reactions, the name will be populated by the
    model based on the order it was added. This could impact seeded simulation 
    results if the order of addition is not preserved.
    """

    def __init__(self, name="", reactants={}, products={}, propensity_function=None, massaction=False, rate=None,
                 annotation=None):
        """
        Initializes the reaction using short-hand notation.
        """

        # Metadata
        self.name = name
        self.annotation = ""

        # We might use this flag in the future to automatically generate
        # the propensity function if set to True.
        if propensity_function is not None:
            self.massaction = False
            self.marate = None
        else:
            self.massaction = True

        self.propensity_function = propensity_function
        self.ode_propensity_function = propensity_function

        if self.propensity_function is not None and self.massaction:
            errmsg = ("Reaction {} You cannot set the propensity type to mass-action and simultaneously set a "
                      "propensity function.").format(self.name)
            raise ReactionError(errmsg)

        self.reactants = {}
        for r in reactants:
            rtype = type(r).__name__
            if rtype == 'instance':
                self.reactants[r.name] = reactants[r]
            else:
                self.reactants[r] = reactants[r]

        self.products = {}
        for p in products:
            rtype = type(p).__name__
            if rtype == 'instance':
                self.products[p.name] = products[p]
            else:
                self.products[p] = products[p]

        if self.massaction:
            self.type = "mass-action"
            if rate is None:
                self.marate = None
            else:
                self.marate = rate
                self.__create_mass_action()
        else:
            self.type = "customized"

            def __customPropParser():
                pow_func = ast.parse("pow", mode="eval").body

                class ExpressionParser(ast.NodeTransformer):
                    def visit_BinOp(self, node):
                        node.left = self.visit(node.left)
                        node.right = self.visit(node.right)
                        if isinstance(node.op, (ast.BitXor, ast.Pow)):
                            # ast.Call calls defined function, args include which nodes
                            # are effected by function call
                            call = ast.Call(func=pow_func,
                                            args=[node.left, node.right],
                                            keywords=[])
                            # Copy_location copies lineno and coloffset attributes
                            # from old node to new node. ast.copy_location(new_node,old_node)
                            call = ast.copy_location(call, node)
                            # Return changed node
                            return call
                        # No modification to node, classes extending NodeTransformer methods
                        # Always return node or value
                        else:
                            return node

                    def visit_Name(self, node):
                        # Visits Name nodes, if the name nodes "id" value is 'e', replace with numerical constant
                        if node.id == 'e':
                            nameToConstant = ast.copy_location(ast.Num(float(np.e), ctx=node.ctx), node)
                            return nameToConstant
                        return node

                expr = self.propensity_function
                expr = expr.replace('^', '**')
                expr = ast.parse(expr, mode='eval')
                expr = ExpressionParser().visit(expr)

                class ToString(ast.NodeVisitor):
                    def __init__(self):
                        self.string = ''

                    def _string_changer(self, addition):
                        self.string += addition

                    def visit_BinOp(self, node):
                        self._string_changer('(')
                        self.visit(node.left)
                        self.visit(node.op)
                        self.visit(node.right)
                        self._string_changer(')')

                    def visit_Name(self, node):
                        self._string_changer(node.id)
                        self.generic_visit(node)

                    def visit_Num(self, node):
                        self._string_changer(str(node.n))
                        self.generic_visit(node)

                    def visit_Call(self, node):
                        self._string_changer(node.func.id + '(')
                        counter = 0
                        for arg in node.args:
                            self.visit(arg)
                            if counter == 0:
                                self._string_changer(',')
                                counter += 1
                        self._string_changer(')')

                    def visit_Add(self, node):
                        self._string_changer('+')
                        self.generic_visit(node)

                    def visit_Div(self, node):
                        self._string_changer('/')
                        self.generic_visit(node)

                    def visit_Mult(self, node):
                        self._string_changer('*')
                        self.generic_visit(node)

                    def visit_UnaryOp(self, node):
                        self._string_changer('(')
                        self.visit_Usub(node)
                        self._string_changer(')')

                    def visit_Sub(self, node):
                        self._string_changer('-')
                        self.generic_visit(node)

                    def visit_Usub(self, node):
                        self._string_changer('-')
                        self.generic_visit(node)

                newFunc = ToString()
                newFunc.visit(expr)
                return newFunc.string
            self.propensity_function = __customPropParser()

    def __str__(self):
        print_string = self.name
        if len(self.reactants):
            print_string += '\n\tReactants'
            for r, stoich in self.reactants.items():
                try:
                    print_string += '\n\t\t' + r.name + ': ' + str(stoich)
                except Exception as e:
                    print_string += '\n\t\t' + r + ': ' + 'INVALID - ' + str(e)
        if len(self.products):
            print_string += '\n\tProducts'
            for p, stoich in self.products.items():
                try:
                    print_string += '\n\t\t' + p.name + ': ' + str(stoich)
                except Exception as e:
                    print_string += '\n\t\t' + p + ': ' + 'INVALID - ' + str(e)
        print_string += '\n\tPropensity Function: ' + self.propensity_function
        return print_string

    def verify(self):
        """ Check if the reaction is properly formatted.
        Does nothing on sucesss, raises and error on failure."""
        if self.marate is None and self.propensity_function is None:
            raise ReactionError("You must specify either a mass-action rate or a propensity function")
        if len(self.reactants) == 0 and len(self.products) == 0:
            raise ReactionError("You must have a non-zero number of reactants or products.")

    def __create_mass_action(self):
        """
        Initializes the mass action propensity function given
        self.reactants and a single parameter value.
        """
        # We support zeroth, first and second order propensities only.
        # There is no theoretical justification for higher order propensities.
        # Users can still create such propensities if they really want to,
        # but should then use a custom propensity.
        total_stoch = 0
        for r in sorted(self.reactants):
            total_stoch += self.reactants[r]
        if total_stoch > 2:
            raise ReactionError("Reaction: A mass-action reaction cannot involve more than two of one species or one "
                                "of two species. To declare a custom propensity, replace 'rate' with "
                                "'propensity_function'.")

        # Case EmptySet -> Y

        propensity_function = self.marate.name
        ode_propensity_function = self.marate.name

        # There are only three ways to get 'total_stoch==2':
        for r in sorted(self.reactants):
            if isinstance(r, str):
                rname = r
            else:
                rname = r.name
            # Case 1: 2X -> Y
            if self.reactants[r] == 2:
                propensity_function = (propensity_function +
                                       "*" + rname + "*(" + rname + "-1)/vol")
                ode_propensity_function += '*' + rname + '*' + rname
            else:
                # Case 3: X1, X2 -> Y;
                propensity_function += "*" + rname
                ode_propensity_function += '*' + rname

        # Set the volume dependency based on order.
        order = len(self.reactants)
        if order == 2:
            propensity_function += "/vol"
        elif order == 0:
            propensity_function += "*vol"

        self.propensity_function = propensity_function
        self.ode_propensity_function = ode_propensity_function

    def setType(self, rxntype):
        """
        Sets reaction type to either "mass-action" or "customized"

        :param rxntype: Either "mass-action" or "customized"
        :type rxntype: str
        """
        if rxntype.lower() not in {'mass-action', 'customized'}:
            raise ReactionError("Invalid reaction type.")
        self.type = rxntype.lower()

        self.massaction = False if self.type == 'customized' else True

    def addReactant(self, S, stoichiometry):
        """
        Adds a reactant to the reaction (species that is consumed)

        :param S: Reactant to add to this reaction.
        :type S: gillespy2.Species
        :param stoichiometry: The stoichiometry of the given reactant.
        :type stoichiometry: int
        """
        if stoichiometry <= 0:
            raise ReactionError("Reaction Stoichiometry must be a \
                                    positive integer.")
        self.reactants[S.name] = stoichiometry

    def addProduct(self, S, stoichiometry):
        """
        Adds a product to the reaction (species that is created)

        :param S: Product to add to this reaction.
        :type S: gillespy2.Species
        :param stoichiometry: The stoichiometry of the given product.
        :type stoichiometry: int
        """
        self.products[S.name] = stoichiometry

    def Annotate(self, annotation):
        """
        Adds a note to the reaction

        :param annotation: An optional note about the reaction.
        :type annotation: str
        """
        self.annotation = annotation

    def sanitized_propensity_function(self, species_mappings, parameter_mappings):
        names = sorted(list(species_mappings.keys()) + list(parameter_mappings.keys()), key=lambda x: len(x),
                       reverse=True)
        replacements = [parameter_mappings[name] if name in parameter_mappings else species_mappings[name]
                        for name in names]
        sanitized_propensity = self.propensity_function
        for id, name in enumerate(names):
            sanitized_propensity = sanitized_propensity.replace(name, "{" + str(id) + "}")
        return sanitized_propensity.format(*replacements)



import gillespy2
from gillespy2.core import gillespyError, GillesPySolver, log
from gillespy2.solvers.utilities import solverutils as cutils
import signal, time #for solver timeout implementation
import os #for getting directories for C++ files
import shutil #for deleting/copying files
import subprocess #For calling make and executing c solver
import inspect #for finding the Gillespy2 module path
import tempfile #for temporary directories
import numpy as np

GILLESPY_PATH = os.path.dirname(inspect.getfile(gillespy2))
GILLESPY_C_DIRECTORY = os.path.join(GILLESPY_PATH, 'solvers/cpp/c_base')
MAKE_FILE = os.path.dirname(os.path.abspath(__file__))+'/c_base/makefile'

def _write_constants(outfile, model, reactions, species, parameter_mappings, resume):
    """
    This function writes the models constants to a user simulation file
    :param outfile: CPP file, used for simulating a model
    :param model: The model that is being simulated
    :param reactions: List of names of a models reactions
    :param species: List of sanitized species names
    :param parameter_mappings: List of sanitized parameter names
    :param resume: If resuming a simulation from a previous one, resume is the results object from the prior simulation.
    Else, it is defaulted to None.
    """

    outfile.write("const double V = {};\n".format(model.volume))
    outfile.write("std :: string s_names[] = {")
    if len(species) > 0:
        # Write model species names.
        for i in range(len(species)-1):
            outfile.write('"{}", '.format(species[i]))
        outfile.write('"{}"'.format(species[-1]))
        outfile.write("};\nunsigned int populations[] = {")
        # Write initial populations.
        for i in range(len(species)-1):
            # If resuming
            if not (resume is None):
                if isinstance(resume, np.ndarray):
                    outfile.write('{}, '.format(int(resume[0][-1][i+1])))
                else:
                    outfile.write('{}, '.format(int(resume[species[i]][-1])))
            else:
                outfile.write('{}, '.format(int(model.listOfSpecies[species[i]].initial_value)))
        if not (resume is None):
            if isinstance(resume, np.ndarray):
                outfile.write('{}'.format(int(resume[0][-1][-1])))
            else:
                outfile.write('{}'.format(int(resume[species[-1]][-1])))
        else:
            outfile.write('{}'.format(int(model.listOfSpecies[species[-1]].initial_value)))
        outfile.write("};\n")
    if len(reactions) > 0:
        # Write reaction names
        outfile.write("std :: string r_names[] = {")
        for i in range(len(reactions)-1):
            outfile.write('"{}", '.format(reactions[i]))
        outfile.write('"{}"'.format(reactions[-1]))
        outfile.write("};\n")
    for param in model.listOfParameters:
        outfile.write("const double {0} = {1};\n".format(parameter_mappings[param], model.listOfParameters[param].value)

                      )


class SSACSolver(GillesPySolver):
    name = "SSACSolver"
    """TODO"""

    def __init__(self, model=None, output_directory=None, delete_directory=True, resume=None):
        super(SSACSolver, self).__init__()
        self.__compiled = False
        self.delete_directory = False
        self.model = model
        self.resume = resume
        if self.model is not None:
            # Create constant, ordered lists for reactions/species/
            self.species_mappings = self.model.sanitized_species_names()
            self.species = list(self.species_mappings.keys())
            self.parameter_mappings = self.model.sanitized_parameter_names()
            self.parameters = list(self.parameter_mappings.keys())
            self.reactions = list(self.model.listOfReactions.keys())

            if isinstance(output_directory, str):
                output_directory = os.path.abspath(output_directory)
            
                if isinstance(output_directory, str):
                    if not os.path.isfile(output_directory):
                        self.output_directory = output_directory
                        self.delete_directory = delete_directory
                        if not os.path.isdir(output_directory):
                            os.makedirs(self.output_directory)
                    else:
                        raise gillespyError.DirectoryError("File exists with the same path as directory.")
            else:
                self.temporary_directory = tempfile.TemporaryDirectory()
                self.output_directory = self.temporary_directory.name
                
            if not os.path.isdir(self.output_directory):
                raise gillespyError.DirectoryError("Errors encountered while setting up directory for Solver C++ files."
                                                   )
            cutils._copy_files(self.output_directory, GILLESPY_C_DIRECTORY)
            self.__write_template()
            self.__compile()

    def __del__(self):
        if self.delete_directory and os.path.isdir(self.output_directory):
            shutil.rmtree(self.output_directory)
        
    def __write_template(self, template_file='SimulationTemplate.cpp'):
        # Open up template file for reading.
        with open(os.path.join(self.output_directory, template_file), 'r') as template:
            # Write simulation C++ file.
            template_keyword = "__DEFINE_"
            # Use same lists of model's species and reactions to maintain order
            with open(os.path.join(self.output_directory, 'UserSimulation.cpp'), 'w') as outfile:
                for line in template:
                    if line.startswith(template_keyword):
                        line = line[len(template_keyword):]
                        if line.startswith("CONSTANTS"):
                            _write_constants(outfile, self.model, self.reactions, self.species, self.parameter_mappings
                                             ,self.resume)
                        if line.startswith("PROPENSITY"):
                            cutils._write_propensity(outfile, self.model, self.species_mappings, self.parameter_mappings
                                                     , self.reactions)
                        if line.startswith("REACTIONS"):
                            cutils._write_reactions(outfile, self.model, self.reactions, self.species)
                    else:
                        outfile.write(line)

    def __compile(self):
        # Use makefile.
        if self.resume:
            if self.resume[0].model != self.model:
                raise gillespyError.ModelError('When resuming, one must not alter the model being resumed.')
            else:
                built = subprocess.run(["make", "-C", self.output_directory, 'UserSimulation'],
                    stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        else:
             try:
                 cleaned = subprocess.run(
                     ["make", "-C", self.output_directory, '-f', MAKE_FILE,
                      'cleanSimulation'],
                     stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                 built = subprocess.run(
                     ["make", "-C", self.output_directory, '-f', MAKE_FILE,
                      'UserSimulation'], stdout=subprocess.PIPE,
                     stderr=subprocess.PIPE)
             except KeyboardInterrupt:
                 log.warning(
                     "Solver has been interrupted during compile time, unexpected behavior may occur.")

        if built.returncode == 0:
            self.__compiled = True
        else:
            raise gillespyError.BuildError("Error encountered while compiling file:\nReturn code: "
                                           "{0}.\nError:\n{1}\n{2}\n".format(built.returncode,
                                                                             built.stdout.decode('utf-8'),
                                                                             built.stderr.decode('utf-8')))

    def get_solver_settings(self):
        """
        :return: Tuple of strings, denoting all keyword argument for this solvers run() method.
        """
        return ('model', 't', 'number_of_trajectories', 'timeout', 'increment', 'seed', 'debug', 'profile')

    def run(self=None, model=None, t=20, number_of_trajectories=1, timeout=0,
            increment=0.05, seed=None, debug=False, profile=False, resume=None, **kwargs):

        pause = False
        if resume is not None:
            if t < resume['time'][-1]:
                raise gillespyError.ExecutionError(
                    "'t' must be greater than previous simulations end time, or set in the run() method as the "
                    "simulations next end time")

        if self is None or self.model is None:
            self = SSACSolver(model, resume=resume)

        if len(kwargs) > 0:
            for key in kwargs:
                log.warning('Unsupported keyword argument to {0} solver: {1}'.format(self.name, key))
        
        unsupported_sbml_features = {
                        'Rate Rules': len(model.listOfRateRules),
                        'Assignment Rules': len(model.listOfAssignmentRules), 
                        'Events': len(model.listOfEvents),
                        'Function Definitions': len(model.listOfFunctionDefinitions)
                        }
        detected_features = []
        for feature, count in unsupported_sbml_features.items():
            if count:
                detected_features.append(feature)

        if len(detected_features):
            raise gillespyError.ModelError('Could not run Model.  SBML Feature: {} not supported by SSACSolver.'
                                           .format(detected_features))

        if self.__compiled:
            self.simulation_data = None
            if resume is not None:
                t = abs(t - resume['time'][-1])

            number_timesteps = int(round(t/increment + 1))
            # Execute simulation.
            args = [os.path.join(self.output_directory, 'UserSimulation'), '-trajectories', str(number_of_trajectories),
                    '-timesteps', str(number_timesteps), '-end', str(t)]
            if seed is not None:
                if isinstance(seed, int):
                    args.append('-seed')
                    args.append(str(seed))
                else:
                    seed_int = int(seed)
                    if seed_int > 0:
                        args.append('-seed')
                        args.append(str(seed_int))
                    else:
                        raise gillespyError.ModelError("seed must be a positive integer")

            # begin subprocess c simulation with timeout (default timeout=0 will not timeout)
            with subprocess.Popen(args, stdout=subprocess.PIPE, start_new_session=True) as simulation:
                return_code = 0
                try:
                    if timeout > 0:
                        stdout, stderr = simulation.communicate(timeout=timeout)
                    else:
                        stdout, stderr = simulation.communicate()
                    return_code = simulation.wait()
                except KeyboardInterrupt:
                    os.killpg(simulation.pid, signal.SIGINT)  # send signal to the process group
                    stdout, stderr = simulation.communicate()
                    pause = True
                    return_code = 33
                except subprocess.TimeoutExpired:
                        os.killpg(simulation.pid, signal.SIGINT)  # send signal to the process group
                        stdout, stderr = simulation.communicate()
                        pause = True
                        return_code = 33

            # Decode from byte, split by comma into array
            stdout = stdout.decode('utf-8').split(',')
            # Parse/return results

            if return_code in [0, 33]:
                trajectory_base, timeStopped = cutils._parse_binary_output(stdout, number_of_trajectories,
                                                                           number_timesteps, len(model.listOfSpecies), stdout,
                                                                           pause=pause)
                if model.tspan[2] - model.tspan[1] == 1:
                    timeStopped = int(timeStopped)
                # Format results
                self.simulation_data = []
                for trajectory in range(number_of_trajectories):
                    data = {'time': trajectory_base[trajectory, :, 0]}
                    for i in range(len(self.species)):
                        data[self.species[i]] = trajectory_base[trajectory, :, i + 1]
                    self.simulation_data.append(data)
            else:
                raise gillespyError.ExecutionError("Error encountered while running simulation C++ file:"
                                                   "\nReturn code: {0}.\nError:\n{1}\n".
                                                   format(simulation.returncode, simulation.stderr))
            if resume is not None or timeStopped != 0:
                self.simulation_data = cutils.c_solver_resume(timeStopped, self.simulation_data, t, resume=resume)

        return self.simulation_data, return_code




import unittest
import tempfile
import os
from gillespy2.core.model import import_SBML
from gillespy2 import ODESolver


class TestSBML(unittest.TestCase):

    def test_sbml_conversion(self):
        try:
            import libsbml
        except ImportError:
            return

        try:
            from urllib2 import urlopen

        except ImportError:
            from urllib.request import urlopen

        sbml_file = 'https://www.ebi.ac.uk/biomodels/model/download/BIOMD0000000028.2?filename=BIOMD0000000028_url.xml'
        response = urlopen(sbml_file)
        tmp = tempfile.NamedTemporaryFile(delete=False)
        tmp.write(response.read())
        tmp.close()
        sbml_model, errors = import_SBML(tmp.name)
        os.remove(tmp.name)
        sbml_results = sbml_model.run(solver=ODESolver)


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
import gillespy2
from example_models import Example
from gillespy2 import ODESolver


class TestBasicODESolver(unittest.TestCase):
    number_of_trajectories = 10

    def test_run_example(self):
        model = Example()
        for i in [1, self.number_of_trajectories]:
            for label in [True, False]:
                with self.subTest(number_of_trajectories=i, show_labels=label):
                    if i > 1:
                        with self.assertLogs(level='WARN'):
                            results = model.run(solver=ODESolver, show_labels=label, number_of_trajectories=i)
                        self.assertEqual(len(results), i)
                    else:
                        results = model.run(solver=ODESolver, show_labels=label, number_of_trajectories=i)

                    if i > 1 or not label:
                        for result in results[1:]:
                            if label:
                                self.assertEqual(results[0], result)

                            else:
                                self.assertTrue(np.array_equal(results[0], result))

    def test_stoich2(self):
        class StoichTestModel(gillespy2.Model):
            def __init__(self, parameter_values=None):
                gillespy2.Model.__init__(self, name='StochTest1')
                A = gillespy2.Species(name='A', initial_value=10)
                B = gillespy2.Species(name='B', initial_value=0)
                self.add_species([A, B])
                k = gillespy2.Parameter(name='k', expression=10)
                self.add_parameter([k])
                r = gillespy2.Reaction(name='r', reactants={A: 2}, products={B:1}, rate=k)
                self.add_reaction([r])
                self.timespan(np.linspace(0, 100, 101))
        model = StoichTestModel()
        result = model.run(solver=ODESolver)
        self.assertAlmostEqual(result['B'][-1], 5, places=3)



if __name__ == '__main__':
    unittest.main()



import unittest
import sys

class TestInit(unittest.TestCase):
    
    def test_sys_init(self):
        x = sys.version_info
        sys.version_info = (2,7)

        with self.assertRaises(Exception):
            Example().run()

        sys.version_info = x

if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
import gillespy2
from gillespy2.core.gillespyError import *
from example_models import Example
from gillespy2 import TauHybridSolver


class TestBasicTauHybridSolver(unittest.TestCase):

    def test_add_rate_rule(self):
        model = Example()
        species = gillespy2.Species('test_species', initial_value=1)
        rule = gillespy2.RateRule(name='rr1',formula='test_species+1',variable='test_species')
        model.add_species([species])
        model.add_rate_rule([rule])
        results = model.run()
        self.assertEqual(results[0].solver_name, 'TauHybridSolver')

    def test_add_rate_rule_dict(self):
        model = Example()
        species2 = gillespy2.Species('test_species2', initial_value=2, mode='continuous')
        species3 = gillespy2.Species('test_species3', initial_value=3, mode='continuous')
        rule2 = gillespy2.RateRule('rule2', species2, 'cos(t)')
        rule3 = gillespy2.RateRule(name='rule3', variable=species3, formula='sin(t)')
        rate_rule_dict = {'rule2': rule2, 'rule3': rule3}
        model.add_species([species2, species3])
        with self.assertRaises(ParameterError):
            model.add_rate_rule(rate_rule_dict)

    def test_add_bad_species_rate_rule_dict(self):
        model = Example()
        rule = gillespy2.RateRule(formula='sin(t)')
        with self.assertRaises(ModelError):
            model.add_rate_rule(rule)

    def test_add_assignment_rule(self):
        model = Example()
        species = gillespy2.Species('test_species4', initial_value=1)
        rule = gillespy2.AssignmentRule(name='ar1', variable=species.name, formula='2')
        model.add_species([species])
        model.add_assignment_rule([rule])
        results = model.run()
        self.assertEquals(results[species.name][0], 2) 
        self.assertEquals(results[species.name][-1], 2)
        self.assertEqual(results[0].solver_name,'TauHybridSolver')

    def test_add_function_definition(self):
        model = Example()
        funcdef = gillespy2.FunctionDefinition(name='fun', function='Sp+1')
        model.add_function_definition(funcdef)
        results = model.run()
        self.assertEqual(results[0].solver_name,'TauHybridSolver')

    def test_add_continuous_species_dependent_event(self):
        model = Example()
        model.listOfSpecies['Sp'].mode = 'continuous'
        eventTrig = gillespy2.EventTrigger(expression='Sp <= 90', initial_value=True, )
        event1 = gillespy2.Event(name='event1', trigger=eventTrig)
        ea1 = gillespy2.EventAssignment(variable='Sp', expression='1000')
        ea2 = gillespy2.EventAssignment(variable='k1', expression='0')
        event1.add_assignment([ea1, ea2])
        model.add_event(event1)
        results = model.run()
        self.assertEqual(results[0].solver_name,'TauHybridSolver')
        self.assertEqual(results['Sp'][-1], 1000)

    def test_add_stochastic_species_dependent_event(self):
        model = Example()
        model.listOfSpecies['Sp'].mode = 'discrete'
        eventTrig = gillespy2.EventTrigger(expression='Sp <= 90', initial_value=True, )
        event1 = gillespy2.Event(name='event1', trigger=eventTrig)
        ea1 = gillespy2.EventAssignment(variable='Sp', expression='1000')
        ea2 = gillespy2.EventAssignment(variable='k1', expression='0')
        event1.add_assignment([ea1, ea2])
        model.add_event(event1)
        results = model.run()
        self.assertEqual(results['Sp'][-1], 1000)
        
    def test_add_continuous_time_dependent_event(self):
        model = Example()
        model.listOfSpecies['Sp'].mode = 'continuous'
        eventTrig = gillespy2.EventTrigger(expression='t >= 10', initial_value=True, )
        event1 = gillespy2.Event(name='event1', trigger=eventTrig)
        ea1 = gillespy2.EventAssignment(variable='Sp', expression='1000')
        ea2 = gillespy2.EventAssignment(variable='k1', expression='0')
        event1.add_assignment([ea1, ea2])
        model.add_event(event1)
        results = model.run()
        self.assertEqual(results['Sp'][-1], 1000)
        
    def test_add_stochastic_time_dependent_event(self):
        model = Example()
        model.listOfSpecies['Sp'].mode = 'discrete'
        eventTrig = gillespy2.EventTrigger(expression='t >= 10', initial_value=True, )
        event1 = gillespy2.Event(name='event1', trigger=eventTrig)
        ea1 = gillespy2.EventAssignment(variable='Sp', expression='1000')
        ea2 = gillespy2.EventAssignment(variable='k1', expression='0')
        event1.add_assignment([ea1, ea2])
        model.add_event(event1)
        results = model.run()
        self.assertEqual(results['Sp'][-1], 1000)
        
    def test_add_param_event(self):
        class EventTestModel(gillespy2.Model):
            def __init__(self):
                gillespy2.Model.__init__(self, name='Event Test Model')
                self.add_species([gillespy2.Species(name='S', initial_value=0)])
                self.add_parameter(gillespy2.Parameter(name='event_tracker',
                                                        expression=99))
                self.add_parameter(gillespy2.Parameter(name='event_tracker2',
                                                        expression=0))
                self.add_reaction(gillespy2.Reaction(name='r1', products={'S':1},
                                                    rate=self.listOfParameters['event_tracker2']))
                eventTrig1 = gillespy2.EventTrigger(expression='t>=2')
                event1 = gillespy2.Event(name='event1', trigger=eventTrig1)
                event1.add_assignment(gillespy2.EventAssignment(
                                        variable='event_tracker', expression='t'))
                eventTrig2 = gillespy2.EventTrigger(expression='t >= event_tracker + 2')
                event2 = gillespy2.Event(name='event2', trigger=eventTrig2)
                event2.add_assignment(gillespy2.EventAssignment(
                                        variable='event_tracker2', expression='t'))
                self.add_event([event1, event2])

        model = EventTestModel()
        results = model.run()
        self.assertGreater(results['S'][-1], 0)

    def test_math_name_overlap(self):
        model = Example()
        gamma = gillespy2.Species('gamma',initial_value=2, mode='continuous')
        model.add_species([gamma])
        k2 = gillespy2.Parameter(name='k2', expression=1)
        model.add_parameter([k2])
        gamma_react = gillespy2.Reaction(name='gamma_react', reactants={'gamma': 1}, products={}, rate=k2)
        model.add_reaction([gamma_react])
        model.run(solver=TauHybridSolver)

    def test_add_bad_expression_rate_rule_dict(self):
        model = Example()
        species2 = gillespy2.Species('test_species2', initial_value=2, mode='continuous')
        rule = gillespy2.RateRule(variable=species2, formula='')
        with self.assertRaises(ModelError):
            model.add_rate_rule(rule)

    def test_ensure_hybrid_dynamic_species(self):
        model = Example()
        species1 = gillespy2.Species('test_species1', initial_value=1,mode='dynamic')
        model.add_species(species1)
        results = model.run()
        self.assertEqual(results[0].solver_name, 'TauHybridSolver')

    def test_ensure_hybrid_continuous_species(self):
        model = Example()
        species1 = gillespy2.Species('test_species1', initial_value=1,mode='continuous')
        model.add_species(species1)
        results = model.run()
        self.assertEqual(results[0].solver_name, 'TauHybridSolver')

    def test_ensure_continuous_dynamic_timeout_warning(self):
        model = Example()
        species1 = gillespy2.Species('test_species1', initial_value=1, mode='dynamic')
        model.add_species(species1)
        with self.assertLogs(level='WARN'):
            results = model.run(timeout=1)


if __name__ == '__main__':
    unittest.main()



import unittest
from example_models import Example
from gillespy2.core import Model, StochMLDocument
from gillespy2.core.gillespyError import *
from gillespy2 import ODESolver
from gillespy2 import NumPySSASolver


class TestStochML(unittest.TestCase):

    def test_StochML_from_and_to_model(self):
        model = Example()
        stochml = StochMLDocument.from_model(model)
        stochml_model = stochml.to_model('model')
        stochml_model.run(solver=ODESolver)
        stochml_model.run(solver=NumPySSASolver)



if __name__ == '__main__':
    unittest.main()



from gillespy2.core import log
try:
    import pyximport
    import numpy as np
    pyximport.install(setup_args={'include_dirs': np.get_include()})
    from gillespy2.solvers.cython.cython_ssa_solver import CythonSSASolver
    can_use_cython = True
    log.debug("Successful Import of Cython solvers.")
except Exception as e:
    can_use_cython = False

__all__ = ['CythonSSASolver'] if can_use_cython else []



import threading
from gillespy2.core import log


class RepeatTimer(threading.Timer):
    """
    Threading timer which repeatedly calls the given function instead of simply ending
    """
    pause = False

    def run(self):
        from IPython.display import clear_output
        type = str.join('', [*self.args[3]])
        self.args = self.args[:3]
        while not self.finished.wait(self.interval):
            self.function(*self.args, **self.kwargs)

        if not self.pause:
            if type == 'progress':
                clear_output()
                print('progress = 100 %')



def display_types():
    return ["graph", "text", "progress"]


def valid_graph_params(live_output_options):
    if live_output_options['type'] not in ['progress', 'graph', 'text']:
        from gillespy2.core.gillespyError import SimulationError
        raise SimulationError("Invalid input to 'live_output', please check spelling and ensure input is"
                              " lower case.")
    if 'interval' not in live_output_options:
        live_output_options['interval'] = 1
    elif live_output_options['interval'] < 0:

        log.warning("In LiveGraphing live_output_options, got \"interval\" = \"{0}\". setting interval = 1"
                    .format(live_output_options['interval']))
        live_output_options['interval'] = 1

    if live_output_options['type'] == "graph" and live_output_options['interval'] < 1:
        log.warning("In LiveGraphing live_output_options, got \"interval\" = \"{0}\". Consider using an interval >= 1 "
                    "when displaying graphs".format(live_output_options['interval']))

    if 'clear_output' not in live_output_options:

        if live_output_options['type'] == "graph" or live_output_options['type'] == "progress":
            live_output_options['clear_output'] = True
        else:
            live_output_options['clear_output'] = False


class LiveDisplayer():
    """
    holds information required for displaying information when live_output = True
    """

    def __init__(self, model=None, timeline=None, number_of_trajectories=1, live_output_options={}, resume=False):

        self.display_type = live_output_options['type']
        self.display_interval = live_output_options['interval']
        self.model = model
        self.resume = resume
        self.timeline = timeline
        self.timeline_len = timeline.size
        self.x_shift = int(timeline[0])
        self.number_of_trajectories = number_of_trajectories
        self.clear_output = live_output_options['clear_output']
        species_mappings = model._listOfSpecies
        self.species = list(species_mappings.keys())
        self.number_species = len(self.species)
        self.current_trajectory = 1
        self.header_printed = False

    def trajectory_header(self):
        return "Trajectory (" + str(self.current_trajectory) + "/" + str(self.number_of_trajectories) + ")"

    def increment_trajectory(self, trajectory_num):
        self.current_trajectory = trajectory_num + 1
        self.header_printed = False

    def print_text_header(self):

        self.header_printed = True
        if self.number_of_trajectories > 1:
            print(self.trajectory_header())

        print("Time      |", end="")
        for species in self.model.listOfSpecies:
            print(species[:10].ljust(10), end="|")
        print("")

    '''
    curr_state and curr_time should be list of len 1 to get reference
    '''
    def display(self, curr_state, curr_time, trajectory_base):
        from IPython.display import clear_output
        from math import floor
        curr_time = curr_time[0]
        curr_state = curr_state[0]

        # necessary for __f function in hybrid solver
        if 't' in curr_state:
            if curr_state['t'] > curr_time:
                curr_time = curr_state['t']
        elif 'time' in curr_state:
            if curr_state['time'] > curr_time:
                curr_time = curr_state['time']

        if self.clear_output:
                clear_output(wait=True)

        if self.display_type == "text":

            if not self.header_printed:
                self.print_text_header()

            print(str(round(curr_time, 2))[:10].ljust(10), end="|")

            for i in range(self.number_species):
                print(str(curr_state[self.species[i]])[:10].ljust(10), end="|")
            print("")

        elif self.display_type == "progress":

            if self.number_of_trajectories > 1:
                print(self.trajectory_header())
            if self.resume is True:
                print("progress =", round(((curr_time-self.x_shift)/self.timeline_len)*100, 2), "%\n"
                      )
            else:
                print("progress =", round((curr_time / (self.timeline_len + self.x_shift)) * 100, 2), "%\n")

        elif self.display_type == "graph":

            import matplotlib.pyplot as plt
            from gillespy2.core.results import common_rgb_values

            entry_count = floor(curr_time) - self.x_shift

            plt.figure(figsize=(18, 10))
            plt.xlim(right=self.timeline[-1])
            plt.xlim(left=self.timeline[0])
            plt.title(self.trajectory_header())
            for i in range(self.number_species):
                line_color = common_rgb_values()[(i) % len(common_rgb_values())]
                plt.plot(trajectory_base[0][:, 0][:entry_count].tolist(),
                         trajectory_base[0][:, i + 1][:entry_count].tolist(), color=line_color,
                         label=self.species[i])

                plt.plot([entry_count - 1, curr_time - self.timeline[0]], [trajectory_base[0][:, i + 1][entry_count - 1]
                    , curr_state[self.species[i]]], linewidth=3,
                         color=line_color)
            plt.legend(loc='upper right')
            plt.show()



from .tau_leaping_solver import TauLeapingSolver


class BasicTauLeapingSolver(TauLeapingSolver):
    from gillespy2.core import log
    log.warning("The name 'BasicTauLeapingSolver' has been deprecated, future versions of GillesPy2 will not allow"
                " this import. Please import 'TauLeapingSolver'")


import unittest
import numpy as np
from example_models import Example
from gillespy2 import TauLeapingSolver
from gillespy2.core import results

class TestBasicTauLeapingSolver(unittest.TestCase):

    model = Example()
    results = model.run(solver=TauLeapingSolver, show_labels=False, number_of_trajectories=1)
    labels_results = model.run(solver=TauLeapingSolver, show_labels=True, number_of_trajectories=1)

    def test_return_type(self):
        assert(isinstance(self.results, np.ndarray))
        assert(isinstance(self.results[0], np.ndarray))
        assert(isinstance(self.results[0][0], np.ndarray))
        assert(isinstance(self.results[0][0][0], np.float))

    def test_return_type_show_labels(self):
        assert(isinstance(self.labels_results, results))
        assert(isinstance(self.labels_results['Sp'], np.ndarray))
        assert(isinstance(self.labels_results['Sp'][0], np.float))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# =============================================================================
# @file    setup.py
# @brief   GillesPy2 setup file
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/GillesPy2/GillesPy2
#
# Note: how to do a PyPI release
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Run the following commands:
#
#   python3 setup.py sdist bdist_wheel
#   twine upload dist/*
#
# =============================================================================

from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install
from setuptools.command.bdist_egg import bdist_egg
from setuptools.command.easy_install import easy_install
import os
from os import path


# Read the contents of auxiliary files.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

SETUP_DIR = path.dirname(os.path.abspath(__file__))

with open(path.join(SETUP_DIR, 'requirements.txt')) as f:
    reqs = f.read().rstrip().splitlines()

with open(path.join(SETUP_DIR, 'README.md'), 'r', errors = 'ignore') as f:
    readme = f.read()

# The following reads the variables without doing an "import handprint",
# because the latter will cause the python execution environment to fail if
# any dependencies are not already installed -- negating most of the reason
# we're using setup() in the first place.  This code avoids eval, for security.

version = {}
with open(path.join(SETUP_DIR, 'gillespy2/__version__.py')) as f:
    text = f.read().rstrip().splitlines()
    vars = [line for line in text if line.startswith('__') and '=' in line]
    for v in vars:
        setting = v.split('=')
        version[setting[0].strip()] = setting[1].strip().replace("'", '')


# Finally, define our namesake.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

setup(name                 = version['__title__'].lower(),
      version              = version['__version__'],
      author               = version['__author__'],
      author_email         = version['__email__'],
      maintainer           = version['__author__'],
      maintainer_email     = version['__email__'],
      license              = version['__license__'],
      url                  = version['__url__'],
      download_url         = version['__download_url__'],
      description          = version['__description__'],
      long_description     = readme,
      long_description_content_type = "text/markdown",
      keywords             = "biochemical simulation, Gillespie algorithm, stochastic simulation, biology",
      project_urls         = {
          "Tracker": "https://github.com/GillesPy2/GillesPy2/issues",
          "Source" : "https://github.com/GillesPy2/GillesPy2",
      },
      packages             = find_packages('.'),
      include_package_data = True,
      install_requires     = reqs,

      classifiers      = [
          'Development Status :: 5 - Production/Stable',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Chemistry',
          'Topic :: Scientific/Engineering :: Mathematics',
          'Topic :: Scientific/Engineering :: Medical Science Apps.',
          'Intended Audience :: Science/Research'
      ],
)



from gillespy2.solvers.cpp.ssa_c_solver import SSACSolver
from gillespy2.solvers.cpp.variable_ssa_c_solver import VariableSSACSolver
from gillespy2.core import log


__all__ = ['SSACSolver', 'VariableSSACSolver']



# Model exceptions
class ModelError(Exception):
    pass


class SpeciesError(ModelError):
    pass


class ReactionError(ModelError):
    pass


class ParameterError(ModelError):
    pass


# Solver specific errors
class SolverError(Exception):
    pass


class DirectoryError(SolverError):
    pass


class BuildError(SolverError):
    pass


class ExecutionError(SolverError):
    pass


class SimulationError(Exception):
    pass


class StochMLImportError(SimulationError):
    pass


class InvalidStochMLError(SimulationError):
    pass


class InvalidModelError(SimulationError):
    pass


class SimulationTimeoutError(SimulationError):
    pass


class EventError(ModelError):
    pass


# Results errors
class ResultsError(Exception):
    pass


class ValidationError(ResultsError):
    pass



from .gillespyError import SimulationError


class GillesPySolver:
    name = "GillesPySolver"
    """ 
    Abstract class for a solver. This is generally called from within a
    gillespy Model through the Model.run function. Returns simulation 
    trajectories.

    :param model: The model on which the solver will operate.
    :type model: gillespy.Model
    
    :param t: The end time of the solver
    :type t: int
    
    :param number_of_trajectories: The number of times to sample the chemical master equation. Each
    trajectory will be returned at the end of the simulation.
    :type number_of_trajectories: int

    :param increment: The time step of the solution
    :type increment: float
   
    :param seed: The random seed for the simulation. Defaults to None. 
    :type seed: int
    
    :param debug: Set to True to provide additional debug information about the     
    simulation.
    :type debug: bool
    
    :param show_labels: Use names of species as index of result object rather than position numbers.
    :type show_labels: bool
    """
    def run(self, model, t=20, number_of_trajectories=1, increment=0.05, seed=None,
            debug=False, profile=False, show_labels=None, **kwargs):
        """ 
        Call out and run the solver. Collect the results.
        """
        raise SimulationError("This abstract solver class cannot be used directly.")

    def get_solver_settings(self):

        raise SimulationError("This abstract solver class cannot be used directly")